/*
 * Performs scan of file (or block device) for sectors looking like NTFS boot
 * sector. Uses 64-bit arithmetic, of cause; thus, potentially may be used for
 * hard disks with more than 4G sectors (>2TB), but NOT tested for such hdds.
 *
 * Vladimir A Smirnov, 2011 (kkatarn@sleepgate.ru, kkatarn@rambler.ru).
 *
 * Static 64-bit linux ELF (compiled under OpenSUSE 11.3) available at:
 * http://ftp.sleepgate.ru/devel/tools/scant/scant.elf64 (581 kb)
 * http://ftp.sleepgate.ru/devel/tools/scant/scant.elf64.xz (217 kb)
 *
 * During my experiments with multiple OSes there was a moment when I had to
 * change active partition on host (Intel X58 - based) RAID1 array made of
 * two 2TB hdds. I was too lazy to start dskprobe.exe and change 2 bits in
 * MBR; instead, I had started WinXP MMC and... successfully trashed content
 * of extended partition (it should be noted - logical disks was created by
 * linux fdisk, not WinXP). DAMN.
 *
 * There is 'gpart' program under Linux. But much better is to do all the
 * restoration by hand. This is for the purpose.
 *
 * Conclusion: if you plan to use MMC, than ALL disk structure must be created
 * by Windows stuff, regardless of how ugly the results look like. Sad.
 *
 * PS. The data was restored )
 */
#define	REPORT_EVERY_MB	200

#define	USE_SEEK64

#ifdef	USE_SEEK64
#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <unistd.h>
#else
#define	_GNU_SOURCE
#include <unistd.h>
#include <sys/syscall.h>

#error	"Debug it first. Than, please, inform kkatarn@sleepgate.ru. Thanks."

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

FILE *in;
FILE *logfile;

unsigned long long offset = 0;

#define		NS	65536
#define		BUFSIZE	(NS * 512)
#define		LOGFILE	"logfile"

#define	NTFS_NTLDR	0
#define	NTFS_BOOTMGR	1
#define	NTFS_STRANGE	2

char *ntfs_typenames[3] = {"NTLDR", "BOOTMGR", "UGLY"};

unsigned char *ptr, buf[BUFSIZE];

char *ntfs_bs_type(unsigned char *ptr)
{
    if(ptr[3] != 'N')
	return NULL;
    if(memcmp(ptr + 3,"NTFS",4) || ptr[510] != 0x55 || ptr[511] != 0xAA)
	return NULL;
    if(!memcmp(ptr + 0x1a2,"NTLDR",5))
	return ntfs_typenames[NTFS_NTLDR];
    else if(!memcmp(ptr + 0x1c0, "BOOTMGR",7))
	return ntfs_typenames[NTFS_BOOTMGR];
    return ntfs_typenames[NTFS_STRANGE];
}

void usage(char *argv0)
{
    printf("Usage: %s <file to be searched for NTFS bootsectors> "
	"[--skip|--skip-mb|--skip-gb <num>]\n", argv0);
}

int main(int argc, char *argv[])
{
    char *st;
    unsigned i, ns_read, mb, mb_prev = 0;
    if(argc != 2 && argc != 4)
    {
	usage(argv[0]);
	return -1;
    }
    in = fopen(argv[1], "rb");
    if(!in)
    {
	printf("Unable to open %s\n", argv[1]);
	return -1;
    }
    logfile = fopen(LOGFILE, "at");
    if(!logfile)
    {
	printf("Unable to open log file '%s'\n", LOGFILE);
	fclose(in);
	return -1;
    }
    if(argc == 4)
    {
	char *s = "s";
	int ret, inno = fileno(in);
	unsigned long long seekto;
	sscanf(argv[3], "%llu", &offset);
	if(!strcasecmp(argv[2],"--skip-mb"))
	    offset = offset << 11;
	else if(!strcasecmp(argv[2],"--skip-gb"))
	    offset = offset << 21;
	else if(strcasecmp(argv[2],"--skip"))
	{
	    usage(argv[0]);
	    fclose(in);
	    fclose(logfile);
	    return -1;
	}
	if(offset == 1)
	    s = "";
	printf("Skipping %llu sector%s (%llu MB, %llu GB): calling ",
	    offset, s, offset >> 11, offset >> 21);
	seekto = offset * 512;
#ifdef	USE_SEEK64
	printf("lseek64(%i, %llu, %u)...", inno, seekto, SEEK_SET);
	ret = lseek64(inno, seekto, SEEK_SET) != (off_t)(-1);
#else
    {
	/*
	 * BUGBUG: there is a problem with OpenSUSE 11.3 headers on my system:
	 * no definition for SYS__llseek. This block of code NOT TESTED.
	 */
	unsigned long hi, lo;
	loff_t res = 0; /* unused */
	lo = (unsigned long)(seekto & 0xFFFFFFFF);
	hi = (unsigned long)((seekto >> 32) & 0xFFFFFFFF);
	printf("linux kernel for _llseek(%i, %lu, %lu, %u)...",
	    inno, hi, lo, SEEK_SET);
#if 0
/* BUGBUG: just a guess, DO NOT enable this! */
#ifndef	SYS__llseek
#define	SYS__llseek	140
#endif
#endif /* 0 */
	ret = syscall(SYS__llseek, inno, hi, lo, &res, SEEK_SET) == 0;
    }
#endif
	if(!ret)
	{
	    printf(" failed!\n");
	    fclose(in);
	    fclose(logfile);
	    return -1;
	}
	printf(" succeeded.\n");
    }
    for(;;offset += NS)
    {
#define	SECTOR_FOUND_FMT "NTFS bootsector of type '%s' found: sector %llu\n"
	mb = (unsigned long)(offset >> 11);
	if(mb - mb_prev > REPORT_EVERY_MB)
	{
	    printf("... still scanning '%s'; offset %u MB (%u GB)\n", argv[1], mb, mb >> 10);
	    mb_prev = mb;
	}
	ns_read = fread(buf, 512, NS, in);
	if(!ns_read)
	{
	    printf("EOF reached reading sector %llu.\n", offset);
	    break;
	}
	for(ptr = buf,i = 0;i < ns_read;i++, ptr += 512)
	{
	    st = ntfs_bs_type(ptr);
	    if(!st)
		continue;
	    printf(SECTOR_FOUND_FMT, st, offset + i);
	    fprintf(logfile,SECTOR_FOUND_FMT, st, offset + i);
	    fflush(logfile);
	    fflush(stdout);
	}
	if(ns_read < NS)
	{
	    printf("EOF reached reading sector %llu.\n", offset + ns_read);
	    break;
	}
    }
    fclose(in);
    fclose(logfile);
    return 0;
}
